from dataclasses import dataclass
from multiprocessing import Pool
from typing import List, NewType, Tuple, Optional, TypedDict

from urlpath import URL
from bs4 import BeautifulSoup, ResultSet, Tag
import requests
from datetime import datetime

from ausbills.types_parliament import House
from ausbills.util.either import Either
from pymonad.either import Left, Right
from pymonad.maybe import Maybe, Just, Nothing

from ausbills import get_logger
from ausbills.util.attr_dict import AttrDict
from ausbills.util.funcs import chunks

DATE_FMT_WA = "%{0}d/%{0}m/%y".format('')  # '#' if sys.platform.startswith('win32') else '-')

PdfUrl = NewType('PdfUrl', str)
UrlStr = NewType('UrlStr', str)

log = get_logger(__file__)

# note: current_bills_url (when I tried it) had 1 bill on the "next page", which is annoying. all_bills_url didn't.
current_bills_url = "https://www.parliament.wa.gov.au/parliament/bills.nsf/" \
                    "screenWebCurrentBills?openform&start=1&count=30000"

all_bills_url = "https://www.parliament.wa.gov.au/parliament/bills.nsf/WebAllBills?openview&start=1&count=30000"
progress_of_bills_url = "https://www.parliament.wa.gov.au/Parliament/Bills.nsf/screenBillsProgress"

waph_base_url = URL("https://www.parliament.wa.gov.au/")


def reqs_get_wa_parli(url, *args, **kwargs):
    return requests.get(url, *args, verify='digicert_sha2_high_assurance_server_ca.pem', **kwargs)


class ScrapeBillProgressException(Exception):
    pass


class ScrapeBillPageException(Exception):
    pass


@dataclass
class BillProgress1House:
    house: House
    fst_read: Maybe[datetime]
    snd_read_hansard: Maybe[Tuple[datetime, str, UrlStr]]
    snd_read: Maybe[datetime]
    consid_detail: Maybe[datetime]
    amend: Maybe[datetime]
    thd_read: Maybe[datetime]

    @classmethod
    def from_tds(cls, house: House, tds: List[Tag]) -> 'BillProgress1House':
        _d = AttrDict(house=house)
        _d.fst_read = BillProgress.parse_date(tds[0].text)
        _d.snd_read_hansard = cls.read_snd_and_hansard(tds[1])
        _d.snd_read = BillProgress.parse_date(tds[2].text)
        _d.consid_detail = BillProgress.parse_date(tds[3].text)
        _d.amend = BillProgress.parse_date(tds[4].text)
        _d.thd_read = BillProgress.parse_date(tds[5].text)
        return BillProgress1House(**_d)

    @staticmethod
    def read_snd_and_hansard(td: Tag) -> Maybe[Tuple[datetime, str, UrlStr]]:
        font_t_children = "__UNSET__"
        try:
            font_t_children = list(td.children.__next__().children)
            if len(font_t_children) < 4:
                return Nothing
            if len(font_t_children) > 4:
                log.error(f"This is very unexpected: {font_t_children}")
            date_str, br, p_dot, hansard_link = font_t_children
            hansard_link: Tag
            date = BillProgress.parse_date(date_str)
            ref = p_dot + hansard_link.text
            url = UrlStr(str(waph_base_url / hansard_link['href']))
            if date.is_just():
                return Just((date.value, ref, url))
        except Exception as e:
            log.warning(f"Exception in read_snd_and_hansard: {e}\n\nfont_t_children: {font_t_children}")
        return Nothing


@dataclass
class BillProgress:
    name: str
    url: UrlStr
    bill_no: str
    la_progress: BillProgress1House
    lc_progress: BillProgress1House
    assent_date: Maybe[datetime]

    @classmethod
    def from_tds(cls, tds: List[Tag]):
        if len(tds) != 15:
            log.warning(f"Got a row in bills progress page that had length /= 15: [ {', '.join(map(str, tds))} ]")
            raise ScrapeBillProgressException(
                f"Got a row in bills progress page that had length /= 15: [ {', '.join(map(str, tds))} ]")
        _d = AttrDict()
        _d.name, _d.url = cls.read_bill_name_link(tds[0])
        _d.url = _d.url if _d.url.startswith('http') else str(waph_base_url / _d.url)
        _d.bill_no = tds[1].text.strip()
        _d.la_progress = BillProgress1House.from_tds(House.LOWER, tds[2:8])
        _d.lc_progress = BillProgress1House.from_tds(House.UPPER, tds[8:14])
        _d.assent_date = cls.parse_date(tds[14].text)
        return BillProgress(**_d)

    @staticmethod
    def read_bill_name_link(td: Tag) -> (str, UrlStr):
        a: Tag = td.children.__next__()
        return a.text, str(waph_base_url / a['href'])

    @staticmethod
    def parse_date(date_text: str) -> Maybe[datetime]:
        _date_text = date_text.strip()
        if _date_text == "":
            return Nothing
        try:
            d_parts = _date_text.split('/')
            extra_day = '0' if len(d_parts[0]) == 1 else ''
            extra_mon = '0' if len(d_parts[1]) == 1 else ''
            d_str = f"{extra_day}{d_parts[0]}/{extra_mon}{d_parts[1]}/{d_parts[2]}"
            dt = datetime.strptime(d_str, DATE_FMT_WA)
            return Just(dt)
        except ScrapeBillPageException as e:
            print(f"This branch never happened during testing. You should see why this prints if it ever does.")
            raise e
            return Nothing


@dataclass
class Bill:
    name: str
    bill_no: str
    url: str
    bill_progress: Maybe[BillProgress]
    # from the detailed page
    synopsis: str
    private_members_bill: Maybe[str]
    status: str
    bill_history: List[Tuple[str, PdfUrl]]
    acts_amended: List[Tuple[str, PdfUrl]]
    related_committee_activity: List[Tuple[str, PdfUrl]]
    lc_supplementary: List[Tuple[str, PdfUrl]]
    messages: List[Tuple[str, PdfUrl]]
    progress_la: List[Tuple[str, datetime]]
    progress_lc: List[Tuple[str, datetime]]
    superseded_lc_supplementary: List[Tuple[str, PdfUrl]]
    comparisons_between_version: Tuple[List[Tuple[str, PdfUrl]], List[Tuple[str, PdfUrl]]]
    conference_of_managers: List[Tuple[str, PdfUrl]]
    notes: List[str]


def scrape_all_bill_progs():
    progress_page_src = reqs_get_wa_parli(progress_of_bills_url).text
    page = BeautifulSoup(progress_page_src, 'lxml')
    prog_table = page.find("table", {'class': "bil_prog_table"})
    rows: ResultSet = prog_table.find("tbody").find_all("tr")
    bill_progs = list()
    for tr_row in rows:
        tds: List[Tag] = list(tr_row.find_all("td"))
        bill_progs.append(BillProgress.from_tds(tds))
    return bill_progs
    # except Exception as e:
    #     continue


def a_tag_to_pair(a_tag: Tag) -> Tuple[str, PdfUrl]:
    label: str = a_tag.text.strip()
    return label, PdfUrl(str(waph_base_url / a_tag['href']))


def li_extract(snd_in_main_pair: Tag) -> List[Tuple[str, PdfUrl]]:
    li_links = snd_in_main_pair.find_all('li', {'class': 'liLink'})
    a_links: List[Tag] = list([li.find('a') for li in li_links])
    r: List[Tuple[str, PdfUrl]] = list(a_tag_to_pair(a) for a in a_links)
    return r


def progress_extract(prog_table: Optional[Tag]) -> List[Tuple[str, datetime]]:
    if prog_table is None:
        return list()
    rows = list(prog_table.find_all('tr'))[1:]
    return list((r[0].text.strip(), datetime.strptime(r[2].text.strip(), "%d %b %Y")) for r in
                (list(row.find_all('td')) for row in rows))


def scrape_bill_from_in_prog(bill_p: BillProgress) -> Tuple[BillProgress, Either[str, Bill]]:
    try:
        _d: TypedDict[Bill] = AttrDict(name=bill_p.name, bill_no=bill_p.bill_no, url=bill_p.url,
                                       bill_progress=Just(bill_p))
        page = BeautifulSoup(reqs_get_wa_parli(bill_p.url).text, 'lxml')
        # name, bill_no, synopsys, status
        heading = page.find('table', {'class': 'billHeading'})
        if heading is None:
            print(heading)
            print(page)
        heading_tds = list(heading.find_all('td'))
        _d.name = heading_tds[0].text.strip()
        _d.bill_no = heading_tds[2].text.strip()
        _d.synopsis = heading_tds[4].text.strip()
        _d.private_members_bill = Just(heading_tds[5].text.strip()) if len(heading_tds) != 7 else Nothing
        _d.status = heading_tds[6 if _d.private_members_bill.is_nothing() else 7].text.strip()
        main_info = heading.find_next_sibling('table')
        top_level_trs = main_info.find_all('tr', recursive=False, )
        main_row_paris = chunks(top_level_trs, 2)
        _d.bill_history = li_extract(main_row_paris.__next__()[1])
        _d.acts_amended = li_extract(main_row_paris.__next__()[1])
        _d.related_committee_activity = li_extract(main_row_paris.__next__()[1])
        _d.lc_supplementary = li_extract(main_row_paris.__next__()[1])
        _d.messages = li_extract(main_row_paris.__next__()[1])
        progress_table = main_row_paris.__next__()[1].find('table')
        la_table = None if progress_table is None else progress_table.find('table', {'class': 'bil_table_LA'})
        lc_table = None if progress_table is None else progress_table.find('table', {'class': 'bil_table_LC'})
        _d.progress_la = progress_extract(la_table)
        _d.progress_lc = progress_extract(lc_table)
        superseded_pair = main_row_paris.__next__()
        _d.superseded_lc_supplementary = li_extract(superseded_pair[1])
        comparison_1: Tag = main_row_paris.__next__()[1]
        comparison_2: Tag = main_row_paris.__next__()[0]
        # we're now offset by 1 to account for an extra row
        _d.comparisons_between_version = (li_extract(comparison_1), li_extract(comparison_2))
        _d.conference_of_managers = li_extract(main_row_paris.__next__()[0])
        if len(_d.conference_of_managers) > 0:
            print(f"found conference_of_managers? {_d.conference_of_managers}")
        _d.notes = list(td.text for td in main_row_paris.__next__()[0].find_all('td'))
        return bill_p, Right(Bill(**_d))
    except Exception as e:
        import traceback
        log.error(f"scrape_bill_from_in_prog ({bill_p.name}) Error: {e}\n\n{traceback.format_tb(e.__traceback__)}")
        return bill_p, Left(str(e))


def scrape_all_bills(bills_progs=None, par_reqs=50) -> List[Tuple[BillProgress, Either[str, Bill]]]:
    bills_progs = bills_progs if bills_progs is not None else scrape_all_bill_progs()
    with Pool(par_reqs) as pool:
        log.info(f"Starting scrape of {len(bills_progs)} bills with {par_reqs} parallel workers.")
        return pool.map(scrape_bill_from_in_prog, bills_progs, par_reqs)



import json

from bs4 import BeautifulSoup, ResultSet
import requests
import datetime

from pymonad.maybe import Maybe, Nothing, Just

from ausbills.json_encoder import AusBillsJsonEncoder
from ausbills.log import get_logger

log = get_logger(__file__)

CHAMBER = "chamber"
SHORT_TITLE = "short_title"
INTRO_HOUSE = "intro_house"
PASSED_HOUSE = "passed_house"
INTRO_SENATE = "intro_senate"
PASSED_SENATE = "passed_senate"
ASSENT_DATE = "assent_date"
URL = "url"
ACT_NO = "act_no"
DOC = "doc"
PDF = "pdf"
HTML = "html"
SUMMARY = "summary"
SPONSOR = "sponsor"
PORTFOLIO = "portfolio"
TEXT_LINK = "text_link"
EM_LINK = "em_link"
ID = "id"
CURRENT_READING = "current_reading"
READINGS = "readings"
bills_legislation_url = "https://www.aph.gov.au/Parliamentary_Business/Bills_Legislation/Bills_Lists/Details_page" \
                        "?blsId=legislation%2fbillslst%2fbillslst_c203aa1c-1876-41a8-bc76-1de328bdb726"


class AllBills(object):
    _bills_data = []
    chambers = ["House", "Senate"]
    this_year = datetime.datetime.now().year

    def __init__(self):
        self._build_dataset()

    def _build_dataset(self):
        try:
            for i in range(2):
                self._scrape_data(i)
        except Exception as e:
            print("Link broken:")
            print(e)

    def _scrape_data(self, table_no):
        markup = requests.get(bills_legislation_url).text
        soup = BeautifulSoup(markup, 'lxml')
        tables = soup.find_all('table')
        trs = tables[table_no].findAll('tr')
        tr = trs.pop(0)
        self.headings = self._get_row_data(tr.findAll('td'))

        for tr in trs:
            row_dict = dict()
            try:
                bill_url_string = str(tr.a['href'])
                row_data = self._get_row_data(tr.findAll('td'))
                row_dict[CHAMBER] = self.chambers[table_no]
                for i in range(len(self.headings)):
                    row_dict[self.headings[i].lower().replace(
                        " ", "_").replace(".", "")] = row_data[i]
                row_dict[URL] = bill_url_string
                row_dict[ID] = bill_url_string.split('?')[-1].split('=')[-1]
                # row_dict[SHORT_TITLE] = row_dict[SHORT_TITLE].replace('\n', '').replace('    ', '').replace(
                #     '\r', '').replace('\u2014\u0080\u0094', ' ').replace('\u00a0', ',').replace('$', 'AUD ')
                row_dict = self._convert_to_datetime(row_dict)
                self._bills_data.append(row_dict)
            except Exception as e:
                print("Bad data", e, ' - ', row_dict[SHORT_TITLE])
                raise e

    def _get_row_data(self, tds):
        row_data = []
        for col in range(7):
            try:
                if tds[col].span:
                    row_data.append(tds[col].span.string)
                else:
                    row_data.append("")
            except Exception as e:
                print(e)
                row_data.append("")
        return row_data

    def _convert_to_datetime(self, bill_dict):
        bill_year = self.this_year

        def to_datetime(indate):
            outdate = None
            if indate is not None:
                if indate != "" and '/' in indate:
                    tempdate = indate.split('/')
                    outdate = datetime.date(
                        bill_year, int(tempdate[1]), int(tempdate[0]))
            return outdate

        for i in range(6):
            year = self.this_year - i
            if str(year) in bill_dict[SHORT_TITLE]:
                bill_year = year
        house_stages = [INTRO_HOUSE, PASSED_HOUSE,
                        INTRO_SENATE, PASSED_SENATE, ASSENT_DATE]
        senate_stages = [INTRO_SENATE, PASSED_SENATE,
                         INTRO_HOUSE, PASSED_HOUSE, ASSENT_DATE]

        for stage in house_stages:
            bill_dict[stage] = to_datetime(bill_dict[stage])

        if bill_dict[CHAMBER] == self.chambers[0]:
            for i in range(len(house_stages) - 1):
                if bill_dict[house_stages[i]] is not None and bill_dict[house_stages[i + 1]] is not None:
                    if bill_dict[house_stages[i]] > bill_dict[house_stages[i + 1]]:
                        d = bill_dict[house_stages[i + 1]]
                        bill_dict[house_stages[i + 1]] = datetime.date(d.year + 1, d.month, d.day)
        elif bill_dict[CHAMBER] == self.chambers[1]:
            for i in range(len(senate_stages) - 1):
                if bill_dict[senate_stages[i]] is not None and bill_dict[senate_stages[i + 1]] is not None:
                    if bill_dict[senate_stages[i]] > bill_dict[senate_stages[i + 1]]:
                        d = bill_dict[senate_stages[i + 1]]
                        bill_dict[senate_stages[i + 1]] = datetime.date(d.year + 1, d.month, d.day)

        return bill_dict

    @property
    def data(self):
        return self._bills_data


_all_bills_global = None


def get_all_bills():
    global _all_bills_global
    if _all_bills_global is None:
        _all_bills_global = AllBills().data
    return _all_bills_global


class Bill:
    _bill_data = dict()

    def __init__(self, bill_dict: dict = None, bill_url: str = None, bill_id: str = None, date_format="YYYY-MM-DD"):
        if all(i is None for i in {bill_id, bill_url, bill_dict}):
            raise ValueError("At least one of the arguments [bill_dict, bill_url, bill_id] must be provided.")

        initial_data = bill_dict
        if initial_data is None:
            try:
                # loop through all bills, if we something matches between the bill's URL or ID and the provided URL or
                # ID, then we have a match. We only care about the first one we find.
                bill = next(filter(
                    lambda b: b is not None,
                    (b if len({b[URL], b[ID]} & {bill_id, bill_url}) != 0 else None for b in get_all_bills())
                ), None)
            except Exception as e:
                log.error(f"Exception finding a matching bill. Exception: {e}")
                raise e
            if bill is None:
                err_source = f'URL: {bill_url}' if bill_url is not None else f'ID: {bill_id}'
                raise ValueError(f"Could not find bill matching {err_source}")
            initial_data = bill

        try:
            self._bill_data = dict(**initial_data)
            self.url = initial_data[URL]
            self.chamber = initial_data[CHAMBER]
            self.short_title = initial_data[SHORT_TITLE]
            self.intro_house = initial_data[INTRO_HOUSE]
            self.passed_house = initial_data[PASSED_HOUSE]
            self.intro_senate = initial_data[INTRO_SENATE]
            self.passed_house = initial_data[PASSED_SENATE]
            self.assent_date = initial_data[ASSENT_DATE]
            self.act_no = initial_data[ACT_NO]
            self.bill_url = requests.get(self.url).text
            self.bill_soup = BeautifulSoup(self.bill_url, 'lxml')
            # date stuff
            self.date_format = date_format
            house_stages = [INTRO_HOUSE, PASSED_HOUSE,
                            INTRO_SENATE, PASSED_SENATE, ASSENT_DATE]
            for stage in house_stages:
                self._bill_data[stage] = self._format_date(self._bill_data[stage])
        except KeyError as e:
            raise KeyError('(class Bill) bill_dict must have all keys. KeyError: ' + str(e))

    def _format_date(self, in_date):
        template = self.date_format
        if in_date is not None and not isinstance(in_date, str):
            out_date = template.replace("YYYY", str(in_date.year))\
                .replace("MM", f"{in_date.month:02d}").replace("DD", f"{in_date.day:02d}")
        else:
            out_date = ''
        return out_date

    def __str__(self):
        return f"<Bill | URL: '{self.url}'>"

    def __repr__(self):
        return ('<{}.{} : {} object at {}>'.format(
            self.__class__.__module__,
            self.__class__.__name__,
            self.url.split('=')[-1],
            hex(id(self))))

    @property
    def summary(self):
        return self.get_bill_summary()

    @property
    def sponsor(self):
        return self.get_sponsor()

    @property
    def portfolio(self):
        return self.get_portfolio()

    @property
    def bill_text_links(self):
        return self.get_bill_text_links()

    @property
    def explanatory_memoranda_links(self):
        return self.get_bill_em_links()

    def get_bill_summary(self):
        try:
            div = self.bill_soup.find("div", id='main_0_summaryPanel')
        except Exception as e:
            div = None
        if div:
            for span_tag in div.find_all('span'):
                span_tag.unwrap()
            return div.p.text
            summary = div.p.text.replace('\n', '').replace('    ', '').replace(
                '\r', '').replace('\u2014\u0080\u0094', ' ').replace('\u00a0', ',').replace('$', 'AUD ')
        else:
            summary = ""
        return summary

    def get_bill_text_links(self):
        empyt_link_dict = {DOC: '',
                           PDF: '',
                           HTML: ''}
        all_texts = []
        tr_code = 'main_0_textOfBillReadingControl_readingItemRepeater_trFirstReading1_'
        for code_n in range(3):
            try:
                tr = self.bill_soup.find(
                    "tr", id=tr_code + str(code_n))
                if tr is None:
                    continue
                links = []
                cells: ResultSet = tr.find_all('td')
                for a in cells[1].find_all('a') if len(cells) > 1 else list():
                    links.append(a['href'])
                if len(links) >= 3:
                    all_texts.append({DOC: links[0],
                                      PDF: links[1],
                                      HTML: links[2]})
            except Exception as e:
                log.warning(f"Exception during get_bill_text_links: {e}")
        reading_dict = {
            'first': empyt_link_dict.copy(),
            'third': empyt_link_dict.copy(),
            'aspassed': empyt_link_dict.copy(),
        }

        for text in all_texts:
            for typ in reading_dict.keys():
                if typ in text[PDF]:
                    reading_dict[typ] = text
        return reading_dict

    def get_bill_em_links(self) -> dict:
        tr = self.bill_soup.find(
            "tr", id='main_0_explanatoryMemorandaControl_readingItemRepeater_trFirstReading1_0')
        if tr is None:
            return dict()
        links = list(tr.find_all('td')[1].find_all('a'))
        return {DOC: links[0],
                PDF: links[1],
                HTML: links[2]}

    def get_sponsor(self) -> Maybe[str]:
        try:
            tr = self.bill_soup.find("div", id='main_0_billSummary_sponsorPanel')
            if tr is None:
                return ''
            return Just(tr.find('dd').text).value
            # return tr.find_all('dd')[0].text.replace('  ', '').replace('\n', '').replace('\r', '')
        except Exception as e:
            log.warning(e)
            return ''
            

    def get_portfolio(self) -> Maybe[str]:
        try:
            tr = self.bill_soup.find("div", id='main_0_billSummary_portfolioPanel')
            return '' if tr is None else Just(tr.find_all('dd')[0].text).value
            # return tr.find_all('dd')[0].text.replace('  ', '').replace('\n', '').replace('\r', '')
        except Exception as e:
            log.warning(e)
            return ''

    @property
    def data(self):
        self._bill_data[CURRENT_READING] = 'first'
        text_type = [DOC, PDF, HTML]
        self._bill_data[SUMMARY] = self.summary
        self._bill_data[SPONSOR] = self.sponsor
        self._bill_data[PORTFOLIO] = self.portfolio
        self._bill_data[READINGS] = self.bill_text_links
        for TEXT in text_type:
            for reading in ['first', 'third', 'aspassed']:
                if self.bill_text_links[reading][TEXT] != '':
                    self._bill_data[TEXT_LINK + '_' + TEXT] = self.bill_text_links[reading][TEXT]
                    self._bill_data[CURRENT_READING] = reading
            self._bill_data[EM_LINK + '_' + TEXT] = self.explanatory_memoranda_links[TEXT]
        return self._bill_data

    def to_json(self) -> str:
        return json.dumps(self._bill_data, cls=AusBillsJsonEncoder)



import base64
import json
from binascii import hexlify
from datetime import datetime

from bs4 import Tag
from pymonad.maybe import Maybe
from dataclasses import is_dataclass

from .log import get_logger
from .types_parliament import House

json_d = json.JSONEncoder.default
log = get_logger(__file__)


class AusBillsJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Maybe):
            return {"$nothing": None} if obj.is_nothing() else {"$just": obj.value}
        if isinstance(obj, Tag):
            return {"$bs4.tag": obj.encode()}
        if isinstance(obj, bytes):
            return {"$bytes": base64.encodebytes(obj).decode()}
        if is_dataclass(obj):
            return dict(**obj.__dict__)
        if isinstance(obj, House):
            return {"$house": obj.value}
        if isinstance(obj, datetime):
            return {"$dateIso8601": obj.isoformat()}
        log.warning(f"Got something of unexpected type ({type(obj)}\n\nObj: {str(obj)}\n\ndir: {dir(obj)}")
        return json_d(self, obj)
